/*
 * Copyright (C) 2012 by
 *   MetraLabs GmbH (MLAB), GERMANY
 * and
 *   Neuroinformatics and Cognitive Robotics Labs (NICR) at TU Ilmenau, GERMANY
 * All rights reserved.
 *
 * Contact: info@mira-project.org
 *
 * Commercial Usage:
 *   Licensees holding valid commercial licenses may use this file in
 *   accordance with the commercial license agreement provided with the
 *   software or, alternatively, in accordance with the terms contained in
 *   a written agreement between you and MLAB or NICR.
 *
 * GNU General Public License Usage:
 *   Alternatively, this file may be used under the terms of the GNU
 *   General Public License version 3.0 as published by the Free Software
 *   Foundation and appearing in the file LICENSE.GPL3 included in the
 *   packaging of this file. Please review the following information to
 *   ensure the GNU General Public License version 3.0 requirements will be
 *   met: http://www.gnu.org/copyleft/gpl.html.
 *   Alternatively you may (at your option) use any later version of the GNU
 *   General Public License if such license has been publicly approved by
 *   MLAB and NICR (or its successors, if any).
 *
 * IN NO EVENT SHALL "MLAB" OR "NICR" BE LIABLE TO ANY PARTY FOR DIRECT,
 * INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
 * THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF "MLAB" OR
 * "NICR" HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * "MLAB" AND "NICR" SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND "MLAB" AND "NICR" HAVE NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR MODIFICATIONS.
 */

/**
 * @file IteratorRangeContainer.h
 *    An iterator range class.
 *
 * @author Erik Einhorn
 * @date   2010/12/03
 */

#ifndef _MIRA_ITERATORRANGECONTAINER_H_
#define _MIRA_ITERATORRANGECONTAINER_H_

#include <list> // to include stl iterator_base

namespace mira {

///////////////////////////////////////////////////////////////////////////////

/**
 * Wraps an STL conform container around a range of values
 * within another container. The Range is given by two
 * iterators.
 * Hence, this container can act as a view to a portion of
 * another container.
 */
template <typename Iterator>
class IteratorRangeContainer
{
public:

	/// Default constructor
	IteratorRangeContainer() :
		mEnd(mBegin) {}

	/// Constructor taking two iterators that define the range
	IteratorRangeContainer(const Iterator& begin, const Iterator& end) :
		mBegin(begin), mEnd(end) {}

public:

	typedef Iterator iterator;
	typedef typename Iterator::reference reference;

public:

	/// Returns an iterator to the first element of the range
	iterator begin() const {
		return mBegin;
	}

	/// Returns an iterator to the end of the range
	iterator end() const  {
		return mEnd;
	}

	/// Returns the first element in the range
	reference front() const { return *begin(); }

	/// Returns the last element in the range
	reference back() const {
		iterator b = end();
		--b;
		return *b;
	}

	/// Returns the number of elements in the given range
	std::size_t size() const {
		return std::distance(begin(), end());
	}

private:

	Iterator mBegin;
	Iterator mEnd;
};

//////////////////////////////////////////////////////////////////////////////

} // namespace

#endif /* _MIRA_ITERATORRANGECONTAINER_H_ */
